from datetime import timedelta

import functools
import numpy.ma as ma
import numpy as np
import matplotlib.pyplot as plt
import os
import pandas as pd
import pathlib

LOG_EPS = -1e150

report_begin_str = (
    "\\documentclass[a4paper]{article} \n \\usepackage[english]{babel} \n \\usepackage[utf8x]"
    "{inputenc} \n \\usepackage[T1]{fontenc} \n \\usepackage[a4paper, top = 3cm, bottom = 2cm, "
    "left = 3cm, right = 3cm, marginparwidth = 1.75cm]{geometry}"
)
report_being_str2 = (
    "\\usepackage{amsmath} \n \\usepackage{amsthm} \n \\usepackage{amsfonts} \n \\usepackage{graphicx} \n \\usepackage[colorinlistoftodos]{todonotes} \n \\usepackage[colorlinks = true, allcolors = blue]{hyperref} \n \\usepackage{multirow} \n \\usepackage{siunitx, etoolbox} \n \\usepackage{subcaption} \n \\usepackage{tikz} \n \\usepackage{standalone} \n \\usetikzlibrary{arrows, automata} \n \\usepackage{listings} \n "
    "\\usepackage{color} \n \\usepackage{textcomp} \n \\theoremstyle{definition} \n \\newtheorem{definition}{Definition} \n \\sisetup{ \n table - align - uncertainty = true, \n separate - uncertainty = true, \n} \n \\renewrobustcmd{\\bfseries}{\\fontseries{b}\\selectfont} \n \\renewrobustcmd {\\boldmath}{} \n \\begin{document}"
)


# https://stackoverflow.com/questions/25389095/python-get-path-of-root-project-structure
def project_root():
    return os.path.dirname(os.path.abspath(__file__))
    return Path(__file__).parent.parent

# TEXT MANIPULATION METHODS #######################
def chop_microseconds(delta):
    return delta - timedelta(microseconds=delta.microseconds)


def convert_to_bold(string):
    return "\033[1m" + string + "\033[0m"


def convert_to_bold_for_latex(string):
    return "\\textbf{" + string + "}"


def remove_digits(s):
    return "".join(i for i in s if not i.isdigit())


def remove_whitespaces_head_and_tail(s):
    if len(s) != 0:
        # remove leading and trailing whitespace
        if s[0] == " ":
            s = s[1:]
        if s[-1] == " ":
            s = s[:-1]
    return s


def set_precision(val, prec=10):
    return round(val, prec)


def llhoods_with_precision(llhoods, prec=40):
    for i in range(len(llhoods)):
        llhoods[i] = set_precision(llhoods[i], prec)
    return llhoods


def contains_all(_str, _list):
    res = True
    for s in _str:
        if s not in _list:
            res = False
            break
    return res
    # return 0 not in [s in list for s in str]


###################### STABLE CALCULATION METHODS #######################
def normalise_safe(xs):
    return xs if np.max(xs) == 0.0 else xs / xs.sum()


def log_sum_probs(log_p1, log_p2):
    log_mx = np.max([log_p1, log_p2])

    return log_mx + np.log(np.exp(log_p1 - log_mx) + np.exp(log_p2 - log_mx))


def log_weighted_sum_normalize_probs(pi_1, log_p1, pi_2, log_p2, pi_3, log_p3):
    x_1 = np.log(pi_1) + log_p1
    x_2 = np.log(pi_2) + log_p2
    x_3 = np.log(pi_3) + log_p3

    xs = [x_1, x_2, x_3]
    log_mx = np.max(xs, axis=0)

    sm = (
        (log_p1 != LOG_EPS) * np.exp(x_1 - log_mx)
        + (log_p2 != LOG_EPS) * np.exp(x_2 - log_mx)
        + (log_p3 != LOG_EPS) * np.exp(x_3 - log_mx)
    )
    return x_1, x_2, x_3, log_mx, sm


def normalize_log_probs(probs):
    reduced_probs = np.exp(probs - max(probs))
    return reduced_probs / reduced_probs.sum()


def logdot(a, b):
    max_a, max_b = np.max(a), np.max(b)
    exp_a, exp_b = np.exp(a - max_a), np.exp(b - max_b)
    c = np.log(np.dot(exp_a, exp_b)) + max_a + max_b

    return c


def ma_multidot(arrays):
    return functools.reduce(ma.dot, arrays)


def multi_logdot(Xs):
    max_Xs = [np.max(X) for X in Xs]
    exp_Xs = [np.exp(X - max_X) for X, max_X in zip(Xs, max_Xs)]
    res = np.linalg.multi_dot(exp_Xs)
    res[np.where(res == 0)] = LOG_EPS
    return np.log(res) + sum(max_Xs)


###############################################################
###################### DATA I/O METHODS #######################
###############################################################
# writing data
def write_data(data, filepath="../../automata/example.dat"):
    f = open(filepath, "w")
    for line in data:
        f.write(str(line) + "\n")
    f.close()


def print_to_file(txt, filename="output.txt"):
    with open(filename, "a+") as f:
        f.write(txt + "\n")


def create_folders(model, _start_over_report):
    """ Creates folders for a column in a dataset
    :param _start_over_report: True/False. True allows creating a new report and False allows appending to the existing report
    :return:
    """
    main_folder = model.config.current_experiment_folder

    # Create inputs, outputs, and results folders
    pathlib.Path(main_folder).mkdir(parents=True, exist_ok=True)
    pathlib.Path(main_folder + "/inputs").mkdir(parents=True, exist_ok=True)
    pathlib.Path(main_folder + "/outputs").mkdir(parents=True, exist_ok=True)
    pathlib.Path(main_folder + "/results").mkdir(parents=True, exist_ok=True)

    # Either creates a new file or appends to an existing one
    if _start_over_report:
        with open(
            model.config.main_experiments_folder
            + "/"
            + model.config.dataset_name
            + "/report.tex",
            "w",
        ) as f:
            # Prints latex file structure
            print(report_begin_str, file=f)
            print(report_being_str2, file=f)

            # Prints title, and section header, etc.
            print(
                "\\title{The Report for the File Named "
                + model.config.dataset_name
                + "} \n \\maketitle",
                file=f,
            )
            print("The number of columns is " + str(model.data.shape[1]) + ".", file=f)
            print("The number of rows is " + str(model.data.shape[0]) + ".", file=f)
            print(
                "\section{Column Name: "
                + model.config.current_column_name.replace("_", "\_")
                + "}",
                file=f,
            )
    else:
        with open(
            model.config.main_experiments_folder
            + "/"
            + model.config.dataset_name
            + "/report.tex",
            "a",
        ) as f:
            print(
                "\section{Column Name: "
                + model.config.current_column_name.replace("_", "\_")
                + "}",
                file=f,
            )


# copies some columns directly
def copy_columns_between_dicts(dict_source, dict_target, columns):
    for column in columns:
        dict_target[column] = dict_source[column]
    return dict_target


###############################################################
#################### VISUALISATION METHODS ####################
###############################################################
def plot_matrix(
    X,
    title="Title",
    xlabel="xlabel",
    ylabel="ylabel",
    figsize=None,
    vmax_=None,
    xticklabels=None,
    yticklabels=None,
    cmap=plt.cm.gray_r,
):
    if figsize is None:
        plt.figure(figsize=(18, 6))
    else:
        plt.figure(figsize=figsize)
    if vmax_ is None:
        VMAX = np.max(X)
    else:
        VMAX = vmax_
    if yticklabels is None:
        pass
    else:
        plt.yticks(range(len(yticklabels)), yticklabels)

    if xticklabels is None:
        pass
    else:
        plt.xticks(range(len(xticklabels)), xticklabels, rotation=90)

    plt.imshow(X, interpolation="none", vmax=VMAX, vmin=0, aspect="auto", cmap=cmap)
    plt.colorbar()
    plt.xlabel(xlabel, fontsize=20)
    plt.ylabel(ylabel, fontsize=20)
    plt.title(title, fontsize=20)
    plt.show()


def plot_normal_type_histogram(x, current_experiment_folder):
    plt.figure()
    plt.hist(np.array(x), bins=20)
    plt.ylabel("data values")
    plt.savefig(current_experiment_folder + "/outputs/histogram.eps", dpi=1000)
    plt.close()


def bar_plot_type_posteriors(
    x, y, current_experiment_folder, _display=False, _save=False
):
    # plots the column posterior in order to show the prediction for column's type
    N = len(y)
    y_pos = np.arange(N)
    plt.figure(figsize=(N * 1.2, 4))
    plt.bar(y_pos, x, align="center", alpha=0.5, width=0.4)
    plt.xticks(y_pos, y)
    plt.ylim(0, 1)
    plt.ylabel("Probability")
    plt.title("Posterior of column type: p(t=k|X)")
    if _save:
        plt.savefig(
            current_experiment_folder + "/outputs/type_posteriors.eps", dpi=1000
        )
    if _display:
        plt.show()
    else:
        plt.close()


# source: http://scipy-cookbook.readthedocs.io/items/Matplotlib_HintonDiagrams.html
def _blob(x, y, area, colour):
    """
    Draws a square-shaped blob with the given area (< 1) at
    the given coordinates.
    """
    hs = np.sqrt(area) / 2
    xcorners = np.array([x - hs, x + hs, x + hs, x - hs])
    ycorners = np.array([y - hs, y - hs, y + hs, y + hs])
    plt.fill(xcorners, ycorners, colour, edgecolor=colour)


# def plot_hinton(W, method=None, _max_value=None, xticklabels=None, yticklabels=None, path=None):
#     """
#     Draws a Hinton diagram for visualizing a weight matrix.
#     Temporarily disables matplotlib interactive mode if it is on,
#     otherwise this takes forever.
#     """
#
#     reenable = False
#     if plt.isinteractive():
#         plt.ioff()
#     plt.clf()
#
#     if reenable:
#         plt.ion()
#
#     special.hinton(W, max_value=_max_value)
#     if xticklabels is not None:
#         plt.xticks(np.arange(len(xticklabels)), xticklabels, rotation=90, fontsize=13)
#     if yticklabels is not None:
#         plt.yticks(np.arange(len(xticklabels)), yticklabels, fontsize=13)
#
#     plt.xlabel('true type', fontsize=17)
#     plt.ylabel('predicted type', fontsize=17)
#
#     if path is not None:
#         plt.savefig(path, dpi=1000, bbox_inches="tight")
#
#     # plt.show()


###############################################################
#################### LATEX METHODS ############################
###############################################################
def print_figure_latex(column_name, f):
    print(
        "\\begin{figure}[!h] \n \\centering \n \\includegraphics[width = \\textwidth]{"
        + column_name
        + "/outputs/type_posteriors.eps} \n \\caption{The posterior probability distribution of the column type for column named "
        + column_name.replace("_", "\\_")
        + ".} \n \\label{fig:"
        + column_name
        + "} \n \\end{figure}",
        file=f,
    )


def print_line_latex(txt, f):
    print(txt, file=f)


def print_row_type_dist_table_latex(
    current_experiment_folder, num_normal_cells, num_missings, num_catch_alls, num_data
):
    table_row_type_dist = pd.DataFrame(
        columns=["", "Number of Entries", "Proportion to the Total Num. of Rows"]
    )

    table_row_type_dist.loc[0] = [
        "Row Type",
        "Number of Entries",
        "Proportion to the Total Num. of Rows",
    ]
    table_row_type_dist.loc[1] = [
        "Normal",
        num_normal_cells,
        round(1.0 * num_normal_cells / num_data, 2),
    ]
    table_row_type_dist.loc[2] = [
        "Missing",
        num_missings,
        round((1.0 * num_missings) / num_data, 2),
    ]
    table_row_type_dist.loc[3] = [
        "Catch-all",
        num_catch_alls,
        round((1.0 * num_catch_alls) / num_data, 2),
    ]

    with open(current_experiment_folder + "/outputs/table_row_type_dist.tex", "w") as f:
        f.write(
            "\\begin{tabular}{|"
            + " | ".join(["c"] * len(table_row_type_dist.columns))
            + "|}\n"
        )
        for i, row in table_row_type_dist.iterrows():
            f.write("\\hline ")
            if i == 0:
                f.write(
                    " & ".join(["\\bfseries " + str(x) for x in row.values]) + " \\\\\n"
                )
            else:
                f.write(" & ".join([str(x) for x in row.values]) + " \\\\\n")
        f.write("\\hline ")
        f.write("\\end{tabular}")


def print_statistics_table_latex(x, current_experiment_folder):
    table_histogram = pd.DataFrame(columns=["Min.", "Max.", "Mean", "Std"])
    table_histogram.loc[0] = ["Min.", "Max.", "Avg.", "Std."]
    table_histogram.loc[1] = [
        round(np.min(x), 2),
        round(np.max(x), 2),
        round(np.mean(x), 2),
        round(np.std(x), 2),
    ]
    with open(
        current_experiment_folder + "/outputs/table_histogram_detail.tex", "w"
    ) as f:
        f.write(
            "\\begin{tabular}{|"
            + " | ".join(["c"] * len(table_histogram.columns))
            + "|}\n"
        )
        for i, row in table_histogram.iterrows():
            f.write("\\hline ")
            if i == 0:
                f.write(
                    " & ".join(["\\bfseries " + str(x) for x in row.values]) + " \\\\\n"
                )
            else:
                f.write(" & ".join([str(x) for x in row.values]) + " \\\\\n")
        f.write("\\hline ")
        f.write("\\end{tabular}")


def print_table_latex(x, current_experiment_folder):
    table_histogram = pd.DataFrame(columns=["Min.", "Max.", "Mean", "Std"])
    table_histogram.loc[0] = ["Min.", "Max.", "Avg.", "Std."]
    table_histogram.loc[1] = [
        round(np.min(x), 2),
        round(np.max(x), 2),
        round(np.mean(x), 2),
        round(np.std(x), 2),
    ]
    with open(
        current_experiment_folder + "/outputs/table_histogram_detail.tex", "w"
    ) as f:
        f.write(
            "\\begin{tabular}{|"
            + " | ".join(["c"] * len(table_histogram.columns))
            + "|}\n"
        )
        for i, row in table_histogram.iterrows():
            f.write("\\hline ")
            if i == 0:
                f.write(
                    " & ".join(["\\bfseries " + str(x) for x in row.values]) + " \\\\\n"
                )
            else:
                f.write(" & ".join([str(x) for x in row.values]) + " \\\\\n")
        f.write("\\hline ")
        f.write("\\end{tabular}")


def evaluate_types(
    _dataset_name, _ptype, _header=None,
):
    dataset_path = "../data/" + _dataset_name + ".csv"
    annotation_path = "../annotations/" + _dataset_name + ".csv"

    df = pd.read_csv(
        dataset_path,
        sep=",",
        encoding="ISO-8859-1",
        dtype=str,
        header=_header,
        keep_default_na=False,
        skipinitialspace=True,
    )
    annotations = pd.read_csv(
        annotation_path,
        sep=",",
        encoding="ISO-8859-1",
        dtype=str,
        keep_default_na=False,
    )

    true_values = annotations["Type"].values.tolist()
    true_values = [true_value.split("-")[0] for true_value in true_values]

    predictions = [col.type for col in _ptype.cols.values()]
    predictions = [
        prediction.replace("date-eu", "date")
        .replace("date-iso-8601", "date")
        .replace("date-non-std-subtype", "date")
        .replace("date-non-std", "date")
        for prediction in predictions
    ]

    column_names = list(_ptype.cols.keys())

    correct_, false_ = 0.0, 0.0
    for i, (prediction, true_value) in enumerate(zip(predictions, true_values)):
        column_name = column_names[i]
        unique_vals, unique_vals_counts = np.unique(
            [str(int_element) for int_element in df[df.columns[i]].tolist()],
            return_counts=True,
        )
        if prediction == true_value:
            correct_ += 1
        else:
            false_ += 1
            print("column name : ", column_names[i])
            indices = _ptype.normal_types[column_name]
            print(
                "\tsome normal data values: ",
                [unique_vals[ind] for ind in indices][:20],
            )
            print("\ttheir counts: ", [unique_vals_counts[ind] for ind in indices][:20])

            indices = _ptype.missing_types[column_name]
            if len(indices) != 0:
                print(
                    "\tsome missing data values: ",
                    [unique_vals[ind] for ind in indices][:20],
                )
                print(
                    "\ttheir counts: ",
                    [unique_vals_counts[ind] for ind in indices][:20],
                )

            indices = _ptype.anomaly_types[column_name]
            if len(indices) != 0:
                print(
                    "\tsome anomalous data values: ",
                    [unique_vals[ind] for ind in indices][:20],
                )
                print(
                    "\ttheir counts: ",
                    [unique_vals_counts[ind] for ind in indices][:20],
                )

            print(
                "\ttrue/annotated type : ",
                true_value,
                "\n\tpredicted type : ",
                prediction,
            )
            print(
                "\tposterior probs: ",
                _ptype.p_t_columns[list(_ptype.p_t_columns.keys())[i]],
            )
            print("\ttypes: ", _ptype.types, "\n")

    print(
        "correct/total = ",
        round(correct_ / len(column_names), 2),
        "(" + str(int(correct_)) + "/" + str(len(column_names)) + ")",
    )



from copy import deepcopy
from scipy import optimize
import numpy as np
from ptype.Column import MISSING_INDEX, ANOMALIES_INDEX
from ptype.Machine import PI
from ptype.utils import (
    LOG_EPS,
    log_weighted_sum_normalize_probs,
    normalize_log_probs,
)


def likelihoods(PI, logP, k):
    _, _, _, log_mx, sm = likelihoods_normalize(PI, logP, k)
    return log_mx + np.log(sm)


def likelihoods_normalize(PI, logP, k):
    return log_weighted_sum_normalize_probs(
        PI[0],
        logP[:, k + LLHOOD_TYPE_START_INDEX],
        PI[1],
        logP[:, MISSING_INDEX - 1],
        PI[2],
        logP[:, ANOMALIES_INDEX - 1],
    )


def sum_weighted_likelihoods(counts_array, logP, k):
    return (counts_array * likelihoods(PI, logP, k)).sum()


def _vecnorm(x, ord=2):
    if ord == np.Inf:
        return np.amax(np.abs(x))
    elif ord == -np.Inf:
        return np.amin(np.abs(x))
    else:
        return np.sum(np.abs(x) ** ord, axis=0) ** (1.0 / ord)


# todo: rename
def wurble(a, b, c):
    return str(a) + "*" + str(b) + "*" + str(c)


LLHOOD_TYPE_START_INDEX = 2


class Trainer:
    """ A Trainer object.

    :param machines: PFSMs to train.
    :param dfs: data frames to train with.
    :param labels: column types labeled by hand, where _label[i][j] denotes the type of j^th column in i^th dataframe.
    """
    def __init__(self, machines, dfs, labels):
        self.machines = machines
        self.dfs = dfs
        self.labels = labels
        self.unique_vals = np.concatenate([np.unique(df.values) for df in dfs])
        self.dfs_unique_vals_counts = Trainer.get_unique_vals_counts(dfs)
        self.machines.set_unique_values(self.unique_vals)

    @staticmethod
    def get_unique_vals_counts(dfs):
        # Finding unique values and their counts
        return {
            i: {
                col: [vs, np.array(counts)]
                for col, (vs, counts) in {
                    col: np.unique(df[col].tolist(), return_counts=True)
                    for col in df.columns
                }.items()
            }
            for i, df in enumerate(dfs)
        }

    def train(self, max_iter=20, uniformly=False, threshold=1e-10):
        """ Train the PFSMs.

        :param max_iter: the maximum number of iterations the optimization algorithm runs as long as it's not converged.
        :param uniformly: a binary variable used to initialize the PFSMs - True allows initializing uniformly rather than using hand-crafted values.
        :param threshold:
        :return:
        """
        if uniformly:
            self.machines.initialize_uniformly()
            self.machines.normalize_params()

        initial = deepcopy(
            self.machines
        )  # shouldn't need this, but too much mutation going on
        training_error = [self.calculate_total_error(self.dfs, self.labels)]

        # Iterates over whole data points
        for n in range(max_iter):
            # Trains machines using all of the training data frames
            self.update_PFSMs()

            # Calculate training and validation error at each iteration
            training_error.append(self.calculate_total_error(self.dfs, self.labels))
            print(training_error)

            if n > 0 and training_error[-2] - training_error[-1] < threshold:
                break

        return initial, self.machines, training_error

    def calculate_total_error(self, dfs, labels):
        all_probs = self.machines.machine_probabilities(self.unique_vals)

        error = 0.0
        for j, (df, df_labels) in enumerate(zip(dfs, labels)):
            for i, column_name in enumerate(list(df.columns)):
                error += self.f_col(all_probs, j, column_name, df_labels[i] - 1)
                # print(
                #     column_name, self.f_col(all_probs, j, column_name, df_labels[i] - 1)
                # )

        return error

    def update_PFSMs(self):
        w_j_z, _ = self.conjugate_gradient(self.machines.get_all_parameters_z())
        self.machines.set_all_probabilities_z(w_j_z)

        for machine in self.machines.forType.values():
            machine.normalize()

    def conjugate_gradient(self, w, J=10, gtol=1e-5):
        d, g = [], []

        gnorm = gtol + 1
        j = 0
        while gnorm > gtol and j < J:
            if j == 0:
                g.append(self.g_cols(w))
                d.append(-g[j])

            res = optimize.line_search(
                self.f_cols, self.g_cols, w, d[j], g[j], self.f_cols(w)
            )
            if res[0] is None:
                return w, j
            else:
                alpha = res[0]
                w = w + alpha * d[j]

                g.append(self.g_cols(w))
                gnorm = _vecnorm(g[j + 1], ord=np.Inf)

                beta_j = max(
                    0, np.dot(g[j + 1].T, g[j + 1] - g[j]) / np.dot(g[j], g[j])
                )  # eq. 7.74 Polak-Ribiere
                d.append(-g[j + 1] + beta_j * d[j])  # eq.7.67

                j += 1

        return w, j

    def f_col(self, all_probs, i, column_name, y_i):
        [temp_x, counts_array] = self.dfs_unique_vals_counts[i][column_name]
        logP = np.array([all_probs[x_i] for x_i in temp_x])
        q = [
            sum_weighted_likelihoods(counts_array, logP, k)
            for k in range(len(self.machines.forType))
        ]

        temp = normalize_log_probs(q)[y_i]
        if temp == 0:
            return 800.0
        else:
            return -np.log(temp)

    def f_cols(self, w_j_z):
        # f: the objective function to minimize. (it is equal to - \sum_{all columns} log p(t=k|X) where k is the correct column type.)
        # Set params: init-transition-final
        self.machines.set_all_probabilities_z(w_j_z)

        # Generate probabilities
        all_probs = self.machines.machine_probabilities(self.unique_vals)

        error = 0.0
        for i, (data_frame, labels) in enumerate(zip(self.dfs, self.labels)):
            for j, column_name in enumerate(list(data_frame.columns)):
                error += self.f_col(all_probs, i, column_name, labels[j] - 1)
        return error

    def do_some_stuff(
        self,
        marginals,
        x_i,
        l,
        temp_g_j,
        state_indices,
        machine,
        alpha,
        t,
        r,
        y_i,
        temp_gra_i,
        counts_array_i,
    ):
        indices_nonzero = np.where(marginals[x_i][l] != 0.0)
        if len(indices_nonzero[0]) != 0:
            q_s = indices_nonzero[0]
            q_primes = indices_nonzero[1]
            for q, q_prime in zip(q_s, q_primes):
                temp_g_j[
                    state_indices[
                        wurble(machine.states[q], alpha, machine.states[q_prime])
                    ]
                ] += self.gradient_transition_marginals(
                    marginals,
                    machine.states[q],
                    alpha,
                    machine.states[q_prime],
                    t,
                    r,
                    str(x_i),
                    y_i,
                    temp_gra_i,
                    counts_array_i,
                )

    def g_col_marginals(self, all_probs, i, col_name, y_i):
        [xs, counts] = self.dfs_unique_vals_counts[i][col_name]
        logP = np.array([all_probs[x] for x in xs])

        # calculates posterior values of types
        r = [
            sum_weighted_likelihoods(counts, logP, k)
            for k in range(len(self.machines.forType))
        ]

        # calculates the gradients for initial, transition, and final probabilities. (note that it is only for non-zero probabilities at the moment.)
        g_j = []
        for t, ty in enumerate(self.machines.types):
            machine = self.machines.forType[ty]
            x_i_indices = np.where(logP[:, t + 2] != LOG_EPS)[0]

            possible_states = [q for q in machine.states if machine.I[q] != LOG_EPS]
            A = likelihoods(PI, logP, t)
            temp_gra = np.exp(PI[0] + logP[:, t + 2] - A)

            # gradient for initial state parameters
            temp_g_j = [
                self.gradient_initial(
                    q,
                    t,
                    xs[x_i_indices],
                    r,
                    temp_gra[x_i_indices],
                    counts[x_i_indices],
                    y_i,
                )
                for q in possible_states
            ]
            g_j = g_j + temp_g_j

            # gradient for transition parameters
            if t == 1:
                marginals = {
                    x_i: np.ones((len(x_i), 1, 1))
                    if p_x_i[t + 2] != LOG_EPS
                    else np.zeros((len(x_i), 1, 1))
                    for x_i, p_x_i in zip(xs, logP)
                }
            else:
                marginals = {
                    x_i: machine.run_forward_backward(str(x_i))
                    if p_x_i[t + 2] != LOG_EPS
                    else np.zeros((len(x_i), len(x_i)))
                    for x_i, p_x_i in zip(xs, logP)
                }
            state_indices = {}
            counter = 0
            temp_g_j = []
            for a in machine.T:
                for b in machine.T[a]:
                    for c in machine.T[a][b]:
                        state_indices[wurble(a, b, c)] = counter
                        temp_g_j.append(0)
                        counter += 1

            for x_i_index, (x_i, temp_gra_i, counts_array_i) in enumerate(
                zip(xs[x_i_indices], temp_gra[x_i_indices], counts[x_i_indices])
            ):
                if logP[x_i_index, t + 2] != LOG_EPS:
                    if t == 1:
                        common_chars = [
                            x for x in machine.alphabet if x in list(str(x_i))
                        ]
                        for alpha in common_chars:
                            common_char_ls = np.where(list(str(x_i)) == alpha)[0]
                            for l in common_char_ls:
                                self.do_some_stuff(
                                    marginals,
                                    x_i,
                                    l,
                                    temp_g_j,
                                    state_indices,
                                    machine,
                                    alpha,
                                    t,
                                    r,
                                    y_i,
                                    temp_gra_i,
                                    counts_array_i,
                                )

                    else:
                        for l, alpha in enumerate(str(x_i)):
                            if alpha in machine.alphabet:
                                self.do_some_stuff(
                                    marginals,
                                    x_i,
                                    l,
                                    temp_g_j,
                                    state_indices,
                                    machine,
                                    alpha,
                                    t,
                                    r,
                                    y_i,
                                    temp_gra_i,
                                    counts_array_i,
                                )
            g_j = g_j + temp_g_j

            # gradient for final-state parameters
            for state in machine.F:
                if machine.F[state] != LOG_EPS:
                    g_j.append(
                        self.gradient_final(
                            state,
                            t,
                            xs[x_i_indices],
                            r,
                            temp_gra[x_i_indices],
                            counts[x_i_indices],
                            y_i,
                        )
                    )

        return -np.array(g_j) / counts.sum()

    def g_cols(self, w_j_z):
        # calculates the gradient vector, i.e. df/dw (=df/dz * dz/dw) where f is the object function to minimize.
        # it returns -g_j because of minimizing instead of maximizing. see the objective function.

        self.machines.set_all_probabilities_z(w_j_z)
        all_probs = self.machines.machine_probabilities(self.unique_vals)

        q_total = None

        for i, (df, labels) in enumerate(zip(self.dfs, self.labels)):
            for j, column_name in enumerate(list(df.columns)):
                if q_total is None:
                    q_total = self.g_col_marginals(
                        all_probs, i, column_name, labels[j] - 1
                    )
                else:
                    q_total += self.g_col_marginals(
                        all_probs, i, column_name, labels[j] - 1
                    )

        return q_total

    ###################### GRADIENT HELPERS #######################
    def scale_wrt_type(self, gradient, q, t, y_i):
        temp = normalize_log_probs(q)[t]
        return gradient * (1 - temp) if t == y_i else -1 * gradient * temp

    def gradient_initial(self, state, t, x, q, temp, counter, y_i):
        machine = self.machines.forType[self.machines.types[t]]
        exp_param = 1 - np.exp(machine.I[state])

        cs = np.array([machine.gradient_initial_state(str(x_i), state) for x_i in x])

        gradient = (temp * counter * cs * exp_param).sum()
        return self.scale_wrt_type(gradient, q, t, y_i)

    def gradient_transition_marginals(
        self, marginals, a, b, c, t, q, x, y_i, temp_gra, counts_array
    ):
        machine = self.machines.forType[self.machines.types[t]]
        temp_mult = (
            temp_gra
            * machine.gradient_abc_new_optimized_marginals(
                marginals[x], str(x), a, b, c
            )
            * counts_array
        )
        exp_param = 1 - np.exp(machine.T[a][b][c])
        gradient = exp_param * temp_mult

        return self.scale_wrt_type(gradient, q, t, y_i)

    def gradient_final(self, final_state, t, x, q, temp, counter, y_i):
        machine = self.machines.forType[self.machines.types[t]]
        exp_param = 1 - np.exp(machine.F[final_state])

        cs = np.array(
            [machine.gradient_final_state(str(x_i), final_state) for x_i in x]
        )
        gradient = sum(temp * counter * cs * exp_param)

        return self.scale_wrt_type(gradient, q, t, y_i)



import numpy as np

import ptype.Machine as Machine
from ptype.Machine import PI

MACHINES = {
    "integer": Machine.Integers(),
    "string": Machine.Strings(),
    "float": Machine.Floats(),
    "boolean": Machine.Booleans(),
    "gender": Machine.Genders(),
    "date-iso-8601": Machine.DateISO_8601(),
    "date-eu": Machine.Date_EU(),
    "date-non-std-subtype": Machine.SubTypeNonstdDate(),
    "date-non-std": Machine.Nonstd_Date(),
    "IPAddress": Machine.IPAddress(),
    "EmailAddress": Machine.EmailAddress(),
}


class Machines:
    def __init__(self, types):
        self.types = types
        self.forType = {t: MACHINES[t] for t in types}
        self.anomalous = Machine.Anomaly()
        self.missing = Machine.Missing()
        self.normalize_params()

    @property
    def machines(self):
        return [self.missing, self.anomalous] + [self.forType[t] for t in self.forType]

    def machine_probabilities(self, col):
        return {v: [m.probability(str(v)) for m in self.machines] for v in col}

    def set_unique_values(self, unique_values):
        for machine in self.machines:
            machine.set_unique_values(unique_values)

    def remove_unique_values(self,):
        for machine in self.machines:
            machine.supported_words = {}

    def update_values(self, unique_values):
        self.remove_unique_values()
        self.set_unique_values(unique_values)

    def normalize_params(self):
        for machine in self.forType.values():
            machine.normalize_params()

    def initialize_uniformly(self):
        for machine in self.forType.values():
            machine.initialize_uniformly()

    def set_all_probabilities_z(self, w_j_z):
        counter = 0
        for machine in self.forType.values():
            counter = machine.set_probabilities_z(counter, w_j_z)

    def get_all_parameters_z(self):
        return [p for machine in self.forType.values() for p in machine.get_parameters_z()]



import joblib
import numpy as np
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import RobustScaler

from pathlib import Path
from ptype import Ptype
from ptype.Column import Column
from ptype.Machines import Machines


class PtypeCat(Ptype.Ptype):
    """The PtypeCat cat object. It uses the following data types: categorical, date, integer, float and string."""

    def __init__(self):
        self.types = [
            "integer",
            "string",
            "float",
            "date-iso-8601",
            "date-eu",
            "date-non-std-subtype",
            "date-non-std",
        ]
        self.machines = Machines(self.types)
        self.verbose = False
        self.lr_clf = joblib.load(Path(__file__).parent.joinpath("LR.sav"))
        self.scaler = joblib.load(Path(__file__).parent.joinpath("scaler.pkl"))


    def _column(self, df, col_name, logP, counts):
        """Returns a Column object for a given data column."""
        col = super()._column(df, col_name, logP, counts)

        # ptype-cat
        t_hat = col.inferred_type()
        if t_hat in ["integer", "string"]:
            feats = col._get_features(counts)
            # magic numbers
            feats[-2:] = self.scaler.transform(feats[-2:].reshape(1, -1))
            ind = np.where(self.lr_clf.classes_ == "categorical")[0][0]
            p_cat = self.lr_clf.predict_proba(feats.reshape(1, -1))[0][ind]
        else:
            p_cat = 0.0
        col.set_p_t_cat(t_hat, p_cat)

        return col




from collections import OrderedDict
from enum import Enum
import joblib
import numpy as np
from ptype.utils import project_root


TYPE_INDEX = 0
MISSING_INDEX = 1
ANOMALIES_INDEX = 2


def _get_unique_vals(col, return_counts=False):
    """List of the unique values found in a column."""
    return np.unique([str(x) for x in col.tolist()], return_counts=return_counts)


# Use same names and values as the constants in Model.py. Could consolidate.
class _Status(Enum):
    TYPE = 1
    MISSING = 2
    ANOMALOUS = 3


class _Feature(Enum):
    U_RATIO = 5
    U_RATIO_CLEAN = 6
    U = 7
    U_CLEAN = 8


class Column:
    def __init__(self, series, p_t, p_z):
        self.series = series
        self.p_t = p_t
        self.p_z = p_z
        self.type = self.inferred_type()
        self.unique_vals, self.unique_vals_counts = _get_unique_vals(
            self.series, return_counts=True
        )
        self._initialise_missing_anomalies()

    def __repr__(self):
        return repr(self.__dict__)

    def _initialise_missing_anomalies(self):
        row_posteriors = self.p_z[self.type]
        max_row_posterior_indices = np.argmax(row_posteriors, axis=1)

        self.normal_indices = list(np.where(max_row_posterior_indices == TYPE_INDEX)[0])
        self.missing_indices = list(
            np.where(max_row_posterior_indices == MISSING_INDEX)[0]
        )
        self.anomalous_indices = list(
            np.where(max_row_posterior_indices == ANOMALIES_INDEX)[0]
        )

    def inferred_type(self):
        """Get most likely inferred type for the column."""
        return max(self.p_t, key=self.p_t.get)

    def get_normal_ratio(self):
        """Get proportion of unique values in the column which are considered neither anomalous nor missing."""
        return round(
            sum(self.unique_vals_counts[self.normal_indices])
            / sum(self.unique_vals_counts),
            2,
        )

    def get_na_ratio(self):
        """Get proportion of unique values in the column which are considered 'missing'."""
        return round(
            sum(self.unique_vals_counts[self.missing_indices])
            / sum(self.unique_vals_counts),
            2,
        )

    def get_an_ratio(self):
        """Get proportion of unique values in the column which are considered 'anomalous'."""
        return round(
            sum(self.unique_vals_counts[self.anomalous_indices])
            / sum(self.unique_vals_counts),
            2,
        )

    def get_normal_values(self):
        """Get list of all values in the column which are considered neither anomalous nor missing."""
        return list(self.unique_vals[self.normal_indices])

    def get_na_values(self):
        """Get a list of the values in the column which are considered 'missing'."""
        return list(self.unique_vals[self.missing_indices])

    def get_an_values(self):
        """Get a list of the values in the column which are considered 'anomalous'."""
        return list(self.unique_vals[self.anomalous_indices])

    def reclassify(self, new_t):
        """Assign a different type to the column, and adjust the interpretation of missing/anomalous values
        accordingly.

        :param new_t: the new type, which must be one of the types known to ptype.
        """
        if new_t not in self.p_z:
            raise Exception(f"Type {new_t} is unknown.")
        self.type = new_t
        self._initialise_missing_anomalies()

    def _get_features(self, counts):
        posterior = OrderedDict()
        # handle equal probs
        for t, p in sorted(self.p_t.items()):
            # aggregate date subtypes
            t_0 = t.split("-")[0]
            if t_0 in posterior.keys():
                posterior[t_0] += p
            else:
                posterior[t_0] = p
        posterior = posterior.values()

        entries = [str(int_element) for int_element in self.series.tolist()]
        U = len(np.unique(entries))
        U_clean = len(self.normal_indices)

        N = len(entries)
        N_clean = sum([counts[index] for index in self.normal_indices])

        u_ratio = U / N
        if U_clean == 0 and N_clean == 0:
            u_ratio_clean = 0.0
        else:
            u_ratio_clean = U_clean / N_clean

        return np.array(list(posterior) + [u_ratio, u_ratio_clean, U, U_clean])

    def set_p_t_cat(self, t_hat, p_cat):
        self.p_t["categorical"] = self.p_t[t_hat] * p_cat
        self.p_t[t_hat] = 1 - self.p_t["categorical"]
        self.type = self.inferred_type()

